Unit AccuDrainDev;

{$mode objfpc}{$H+}

Interface

Uses
  Classes, SysUtils,
  BaseUnix,Errors,
  USBUtil, EZUSB,
  AccuDrainTypes;

Type
  EAccuDrain                    = class(Exception);
  EAccuDrainVanished            = class(EAccuDrain);
  EAccuDrainUSBComm             = class(EAccuDrain);

  TSetupDataNotifier = Procedure(SetupData:TSetupData) of object;

  { TAccuDrainDev }

  TAccuDrainDev = class(TUSBDeviceWithFirmware)
  private
    FChannelDataArray      : TThreadList;
    FSetupDataArray        : TThreadList;
    FDataAvailableEvent    : PRTLEvent;
    FOnSetupDataAdded      : TSetupDataNotifier;
    FOnSetupDataSent       : TSetupDataNotifier;
    FCurrentSetup          : TSetupData;
    Procedure HandleUSBResult(Result: Integer);
    Procedure Send(Const SetupData:TSetupData);
    Procedure Add(Const ChannelData:TAllChannelsDataStored);
    Procedure Add(Const SetupData:TSetupData);
    Function  Get(Out SetupData:TSetupData) : Cardinal;
  public
    Constructor Create(idVendor,idProduct,idVendorUnconf,idProductUnconf:LongInt;
                       AFirmwareFilename:String);
    Destructor  Destroy; override;
    Procedure Recv;
    Procedure Send;
    Function  Get(Out ChannelData:TAllChannelsDataStored) : Cardinal;
    Procedure SetChannel(AChannel:WhichChannelEnum;AChannelSetup:TChannelSetupData);
    Procedure SetChannels(AChannels:WhichChannelSet;AChannelSetup:TAllChannelsSetupData);
    Procedure SetDivisors(ADivLedsStart:Byte;ADivMeasureStart:Byte;ADivStoreStart:Byte;ADivPauseStart:UInt16;ADivPauseStop:Byte);
    Procedure SetUBattPresent(AUBattPresent:UInt16);
    Procedure SetRegFactor(ARegFactor:UInt16);

    property DataAvailableEvent : PRTLEvent Read FDataAvailableEvent Write FDataAvailableEvent;
    property OnSetupDataAdded   : TSetupDataNotifier Read FOnSetupDataAdded Write FOnSetupDataAdded;
    property OnSetupDataSent    : TSetupDataNotifier Read FOnSetupDataSent  Write FOnSetupDataSent;
    property CurrentSetup       : TSetupData read FCurrentSetup;
  End;

Implementation

{ TAccuDrainDev }

Constructor TAccuDrainDev.Create(idVendor, idProduct, idVendorUnconf,
  idProductUnconf: LongInt;
  AFirmwareFilename: String);
Begin
  inherited Create(idVendor,idProduct,idVendorUnconf,idProductUnconf,
    TUSBDeviceEZUSB,AFirmwareFilename,
    1,0,1, 1);    { <-- TODO }
  FChannelDataArray   := TThreadList.Create;
  FSetupDataArray     := TThreadList.Create;
End;

Destructor TAccuDrainDev.Destroy;
Begin
  try
    FChannelDataArray.Free;
    FSetupDataArray.  Free;
    Inherited Destroy;
  finally
  End;
End;

Procedure TAccuDrainDev.HandleUSBResult(Result:Integer);
Begin
  Case -Result of
    ESysENODEV : raise EAccuDrainVanished.Create('Device has vanished');
  else
    raise EAccuDrainUSBComm.Create('USB problem: '+StrError(-Result));
  End;
End;

Procedure TAccuDrainDev.Send(Const SetupData: TSetupData);
Var Ret : Integer;
Begin
  { send setup data }
  Ret := inherited Send(2,SetupData,sizeof(SetupData),200);
  { check result }
  if Ret < 0 then
    HandleUSBResult(Ret)
  else if Ret <> sizeof(SetupData) then
    raise EAccuDrainUSBComm.Create('Could only send '+IntToStr(Ret)+' bytes of '+IntToStr(sizeof(SetupData))+' bytes SetupData');
  { store as current setup because we can't ask the device for it }
  StoreSetup(SetupData,FCurrentSetup);
  { notify }
  if assigned(FOnSetupDataSent) then
    FOnSetupDataSent(SetupData);
End;

Procedure TAccuDrainDev.Recv;
Var ChannelData : Array[0..1] of TAllChannelsDataStored;
    Ret         : Integer;
Begin
  Ret := inherited Recv(2,ChannelData,sizeof(ChannelData),95);   // a little faster than 10Hz
  if Ret = -ESysETIMEDOUT then
    // WriteLn('No data, timed out')  // ignore that
  else if Ret < 0 then
    HandleUSBResult(Ret)
  else if Ret = 0 then
    // WriteLn('No data, zero size reply')  // ignore that
  else if Ret = sizeof(ChannelData[0]) then
    Begin
      Add(ChannelData[0]);
    End
  else if Ret = sizeof(ChannelData) then
    Begin
      Add(ChannelData[0]);
      Add(ChannelData[1]);
    End
  else
    raise EAccuDrainUSBComm.Create('Could only receive '+IntToStr(Ret)+' bytes');
End;

Procedure TAccuDrainDev.Send;
Var SetupData : TSetupData;
    I         : Cardinal;
Begin
  While true do
    Begin
      I := Get(SetupData);
      if I = 0 then Break;
      Send(SetupData);
      if I > 1 then
        Sleep(10);  // give the device some time to handle the request
    End;
End;

Procedure TAccuDrainDev.Add(Const ChannelData: TAllChannelsDataStored);
Var D : PAllChannelsDataStored;
Begin
  New(D);
  D^ := ChannelData;
  FChannelDataArray.Add(D);
  if FDataAvailableEvent <> Nil then
    RTLEventSetEvent(FDataAvailableEvent);
End;

Function TAccuDrainDev.Get(Out ChannelData: TAllChannelsDataStored): Cardinal;
Var List : TList;
Begin
  List := FChannelDataArray.LockList;
  try
    Result := List.Count;
    if Result > 0 then
      Begin
        ChannelData := PAllChannelsDataStored(List.Items[0])^;
        Dispose(PAllChannelsDataStored(List.Items[0]));
        List.Delete(0);
      End;
  finally
    FChannelDataArray.UnlockList;
  End;
End;

Procedure TAccuDrainDev.Add(Const SetupData: TSetupData);
Var D : PSetupData;
Begin
  New(D);
  D^ := SetupData;
  FSetupDataArray.Add(D);
  if assigned(FOnSetupDataAdded) then
    FOnSetupDataAdded(SetupData);
End;

Function TAccuDrainDev.Get(Out SetupData: TSetupData): Cardinal;
Var List : TList;
Begin
  List := FSetupDataArray.LockList;
  try
    Result := List.Count;
    if Result = 0 then
      Exit;
    SetupData := PSetupData(List.Items[0])^;
    Dispose(PSetupData(List.Items[0]));
    List.Delete(0);
  finally
    FSetupDataArray.UnlockList;
  End;
End;

Procedure TAccuDrainDev.SetChannel(AChannel: WhichChannelEnum;AChannelSetup: TChannelSetupData);
Var SetupData : TSetupData;
Begin
  if not (AChannel in [wcCh0,wcCh1,wcCh2,wcCh3]) then
    raise Exception.Create('TAccuDrainDev.SetChannel used with invalid channel argument');
  SetupData.AllChannelSetupData[ChannelIndex(AChannel)] := AChannelSetup;
  SetupData.WhichChannels := WhichChannel([AChannel]);
  Add(SetupData);
End;

Procedure TAccuDrainDev.SetChannels(AChannels: WhichChannelSet;AChannelSetup: TAllChannelsSetupData);
Var SetupData : TSetupData;
Begin
  SetupData.AllChannelSetupData := AChannelSetup;
  SetupData.WhichChannels       := WhichChannel(AChannels * [wcCh0,wcCh1,wcCh2,wcCh3]);
  Add(SetupData);
End;

Procedure TAccuDrainDev.SetDivisors(ADivLedsStart: Byte; ADivMeasureStart: Byte;ADivStoreStart: Byte; ADivPauseStart: UInt16;ADivPauseStop: Byte);
Var SetupData : TSetupData;
Begin
  SetupData.DivLedsStart    := ADivLedsStart;
  SetupData.DivMeasureStart := ADivMeasureStart;
  SetupData.DivStoreStart   := ADivStoreStart;
  SetupData.DivPauseStart   := ADivPauseStart;
  SetupData.DivPauseStop    := ADivPauseStop;
  SetupData.WhichChannels   := WhichChannel([wcDiv]);
  Add(SetupData);
End;

Procedure TAccuDrainDev.SetUBattPresent(AUBattPresent: UInt16);
Var SetupData : TSetupData;
Begin
  SetupData.UBattPresent  := AUBattPresent;
  SetupData.WhichChannels := WhichChannel([wcUBattPresent]);
  Add(SetupData);
End;

Procedure TAccuDrainDev.SetRegFactor(ARegFactor: UInt16);
Var SetupData : TSetupData;
Begin
  SetupData.RegFactor     := ARegFactor;
  SetupData.WhichChannels := WhichChannel([wcRegFactor]);
  Add(SetupData);
End;

End.
