(*****************************************************************************
 *      Copyright (C) 2003  Johann Glaser (Johann.Glaser@gmx.at)
 *
 *      This file is part of the Bar Controller
 * 
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *****************************************************************************)

{$MODE OBJFPC}
Unit EZUSB;

Interface
Uses LibUSB,USBUtil,IntelHex;

Type
  TUSBDeviceEZUSB = class(TUSBDeviceWithoutFirmware)
  protected
    Procedure   Download(AFilename:String); override; { IHX file }
  public
    Constructor Create(idVendor,idProduct:LongInt;ARetries:Longint=0); override; overload;
  End;


Function  EZUSB_LoadMem  (Dev:PUSBDevHandle;HexRecord:PIntelHexRecord):LongInt;
Procedure EZUSB_Configure(Devs:PUSBDevice;IHXFileName:ShortString);

Implementation
Uses SysUtils,Errors;

{ Vendor specific request code for Anchor Upload/Download.
  This one is implemented in the core. }
Const ANCHOR_LOAD_INTERNAL = $A0;
{ EZ-USB Control and Status Register.  Bit 0 controls 8051 reset }
      CPUCS_REG            = $7F92;

(*****************************************************************************
  EZUSB_WriteMem: write "len" bytes from "*data" at memory position "pos" to
    the previously usb_open'ed USB device usbdev.
  Errors: opening the file: -errno
          allocating memory: returns -ENOMEM
  returns: >0 .... count of lines read
           <0 .... error number
 *****************************************************************************)
Function EZUSB_WriteMem(Dev:PUSBDevHandle;Pos:LongInt;Var Data;Len:LongInt) : LongInt;
Begin
//  WriteLn('    EZUSB WriteMem Pos = $',IntToHex(Pos,4),', Len = ',Len);
  Result := usb_control_msg(Dev,$40,ANCHOR_LOAD_INTERNAL,Pos,0,@Data,Len,300);
End;
(* --------------------------------------------------------------------- *)
Function EZUSB_8051_Reset(Dev:PUSBDevHandle;ResetBit:Byte) : LongInt;
Begin
//  WriteLn('  EZUSB Reset: Bit = ',ResetBit);
  Result := EZUSB_WriteMem(Dev,CPUCS_REG,ResetBit,1);
End;
(* --------------------------------------------------------------------- *)
Function EZUSB_LoadMem(Dev:PUSBDevHandle;HexRecord:PIntelHexRecord):LongInt;
Begin
  Result := EZUSB_8051_Reset(Dev,1);
  While (HexRecord <> Nil) and (HexRecord^.TheType = 0) do
    Begin
      Result := EZUSB_WriteMem (Dev,HexRecord^.Address,HexRecord^.Data,HexRecord^.Length);
      if Result < 0 then
        raise EInOutError.CreateFmt('EZUSB_WriteMem failed (%d %04X %p %d) with error %d: %s\n',[Result,HexRecord^.Address,@HexRecord^.Data,HexRecord^.Length,usb_error_errno,StrError(usb_error_errno)]);
      HexRecord := HexRecord^.Next;
    End;
  Result := EZUSB_8051_Reset(Dev,0);
End;

Procedure EZUSB_Configure(Devs:PUSBDevice;IHXFileName:ShortString);
Var Handle   : PUSBDevHandle;
    Firmware : PIntelHexRecord;
Begin
  if Devs = Nil then Exit;
  Firmware := ReadHexFile(IHXFileName); { might raise an exception if the ihx file is currupt }
  if Firmware = Nil then
    Begin
      WriteLn('ERROR: No firmware found!');
      Exit;
    End;
  try
    While Devs <> Nil do
      Begin
        try
          Handle := USBOpenAndSetup(Devs,1,0,0,5);
        except
          on E : Exception do
            Begin
              WriteLn('Couldn''t open the USB Device '+StrPas(Devs^.Filename)+' for firmware download: '+E.Message);
              Devs := Devs^.Next;
              Continue;
            End;
        End;
//        WriteLn('Configuring USB device ',Devs^.Filename);
        try
          EZUSB_LoadMem(Handle,Firmware); { might raise an exception on usb errors }
          usb_close(Handle);
        except
          { be error tolerant and don't cancel configuring several devices if one breaks }
          on E : Exception do
            WriteLn('  Error: '+E.Message);
        End;
        Devs := Devs^.Next;
      End;
  finally
    FreeIntelHex(Firmware);
  End;
End;

{ TUSBDeviceEZUSB }

Constructor TUSBDeviceEZUSB.Create(idVendor, idProduct: LongInt; ARetries: Longint);
Begin
  { TODO: remove: TUSBDevice(Self).}Create(idVendor,idProduct,1,0,0,ARetries);
  { TODO: strange, this gives an Access exception if inside of Create an exception is raised }
End;

Procedure TUSBDeviceEZUSB.Download(AFilename: String);
Var Firmware : PIntelHexRecord;
Begin
  Firmware := ReadHexFile(AFileName); { might raise an exception if the ihx file is currupt }
  try
    EZUSB_LoadMem(FHandle,Firmware);  { might raise an exception on usb errors }
  finally
    FreeIntelHex(Firmware);
  End;
End;


End.

